package BSS.Controllers;

import BSS.*;

import java.util.Map;

public abstract class BaseController implements Controller, EnvironmentCodes
{
    private int _id;
    private Logger _logger;
    private Device _dev;

    //--- Public section ---
    public BaseController(int id, Logger logger, Device device)
    {
        if (null == logger)
        {
            throw new NullPointerException("Null logger.");
        }
        if (null == device)
        {
            throw new NullPointerException("Null device.");
        }
        _id = id;
        _dev = device;
        _logger = logger;
    }

    protected int ID()
    {
        return _id;
    }

    public String toString()
    {
        return "Controller#" + _id;
    }

    //--- Controller interface methods ---
    public abstract void init();

    public abstract void run();

    public void envOnEvent(int event_type) {}

    public Message msgOnReceive(Message msg) { return null; }

    public void smsOnReceive(SMS sms) {}

    //--- Logging methods ---
    protected void log(int level, String message)
    {
        _logger.log(level, message);
    }

    protected void log_start(String automaton_name, int state, int event)
    {
        _logger.log(2, "Start automaton " + automaton_name + "#" + _id + " state:" + state + " event:" + event);
    }

    protected void log_end(String automaton_name, int state)
    {
        _logger.log(2, "End automaton " + automaton_name + "#" + _id + " state:" + state);
    }

    protected void log_x(String x_name, boolean result, String x_desc)
    {
        _logger.log(2, "Condition " + x_name + " is " + result + " at " + toString() + " (" + x_desc + ")");
    }

    protected void log_z(String z_name, String z_desc)
    {
        _logger.log(2, "Action " + z_name + " at " + toString() + " (" + z_desc + ")");
    }

    //--- Device interaction methods ---
    protected Message msgMakeReply(Message msg, int data)
    {
        return new Message(_id, msg.src_id, 0, data);
    }

    protected void smsSend(String to, int object_id, int event)
    {
        _dev.sendSMS(new SMS(to, "BSS", object_id, event));
    }

    protected Message msgSend(int to_id, int event_type)
    {
        return _dev.sendMessage(new Message(to_id, _id, event_type));
    }

    protected Message msgSend(int to_id, int event_type, int event_data)
    {
        return _dev.sendMessage(new Message(to_id, _id, event_type, event_data));
    }

    protected boolean msgRoomOnGuard(int room_id)
    {
        Message resp = msgSend(room_id, REQ_ROOM_ON_GUARD);
        return (null == resp) ? false : (1 == resp.data);
    }

    protected boolean msgRoomHasPeople(int room_id)
    {
        Message resp = msgSend(room_id, REQ_ROOM_HAS_PEOPLE);
        return (null == resp) ? false : (1 == resp.data);
    }

    protected boolean msgEntryIsLocked(int entry_id)
    {
        Message resp = msgSend(entry_id, REQ_ENTRY_IS_LOCKED);
        return (null == resp) ? false : (1 == resp.data);
    }

    protected boolean msgEntryIsClosed(int entry_id)
    {
        Message resp = msgSend(entry_id, REQ_ENTRY_IS_CLOSED);
        return (null == resp) ? false : (1 == resp.data);
    }

    //--- Device interaction methods ---
    protected Map sysLoadConfig()
    {
        return _dev.loadConfiguration();
    }

    protected boolean envHasContact(int device_id)
    {
        return 0 != (_dev.getStatus(DEV_CONTACT_SENSOR, device_id) & 0x01);
    }

    protected boolean envSmartCardInserted(int device_id)
    {
        return 0 != (_dev.getStatus(DEV_SMARTCARD_DRIVE, device_id) & 0x01);
    }

    protected boolean envSmartCardRead(int device_id)
    {
        return 0 != (_dev.getStatus(DEV_SMARTCARD_DRIVE, device_id) & 0x02);
    }

    protected void envOn_LED(int device_id)
    {
        _dev.setStatus(DEV_LED, device_id, (byte)0x01);
    }

    protected void envOff_LED(int device_id)
    {
        _dev.setStatus(DEV_LED, device_id, (byte)0x02);
    }

    protected void envLock(int device_id)
    {
        _dev.setStatus(DEV_LOCK, device_id, (byte)0x01);
    }

    protected void envUnlock(int device_id)
    {
        _dev.setStatus(DEV_LOCK, device_id, (byte)0x02);
    }

    protected String envReadSmartCard(int device_id)
    {
        return new String(_dev.readData(DEV_SMARTCARD_DRIVE, device_id));
    }

    protected void envResetSmartCardDrive(int device_id)
    {
        _dev.setStatus(DEV_SMARTCARD_DRIVE, device_id, (byte)0x01);
    }

    protected boolean envPhotoDark(int device_id)
    {
        return 0 != (_dev.getStatus(DEV_PHOTO_SENSOR, device_id) & 0x01);
    }

    protected void envStopGate(int device_id)
    {
        _dev.setStatus(DEV_MOTOR, device_id, (byte)0x04);
    }

    protected void envCloseGate(int device_id)
    {
        _dev.setStatus(DEV_MOTOR, device_id, (byte)0x02);
    }

    protected void envOpenGate(int device_id)
    {
        _dev.setStatus(DEV_MOTOR, device_id, (byte)0x01);
    }

    protected void envEnergize(int device_id)
    {
        _dev.setStatus(DEV_ENERGIZER, device_id, (byte)0x01);
    }

    protected void envUnenergize(int device_id)
    {
        _dev.setStatus(DEV_ENERGIZER, device_id, (byte)0x02);
    }

    protected boolean envHasMetal(int device_id)
    {
        return 0 != (_dev.getStatus(DEV_METAL_SENSOR, device_id) & 0x01);
    }

    protected String envReadKeyboard(int device_id)
    {
        return new String(_dev.readData(DEV_KEYBOARD, device_id));
    }

    protected void envPrint(int device_id, String text)
    {
        _dev.writeData(DEV_PRINTER, device_id, text.getBytes());
    }

    protected boolean envHotAndSmoky(int device_id)
    {
        return 0 != (_dev.getStatus(DEV_FIRE_SENSOR, device_id) & 0x01);
    }

    protected void envSmokeOn(int device_id)
    {
        _dev.setStatus(DEV_SMUDGE, device_id, (byte)0x01);
    }

    protected void envSmokeOff(int device_id)
    {
        _dev.setStatus(DEV_SMUDGE, device_id, (byte)0x02);
    }

    protected void envCameraOn(int device_id)
    {
        _dev.setStatus(DEV_CAMERA, device_id, (byte)0x01);
    }

    protected void envCameraOnAndAutoOff(int device_id)
    {
        _dev.setStatus(DEV_CAMERA, device_id, (byte)0x04);
    }

    protected void envCameraOff(int device_id)
    {
        _dev.setStatus(DEV_CAMERA, device_id, (byte)0x02);
    }

    protected void envInterferenceOn(int device_id)
    {
        _dev.setStatus(DEV_INTERFERENCE, device_id, (byte)0x01);
    }

    protected void envInterferenceOff(int device_id)
    {
        _dev.setStatus(DEV_INTERFERENCE, device_id, (byte)0x02);
    }

    protected void envStopLift(int device_id)
    {
        _dev.setStatus(DEV_MOTOR, device_id, (byte)0x04);
    }

    protected void envFireExtinguisherOn(int device_id)
    {
        _dev.setStatus(DEV_FIRE_EXTINGUISHER, device_id, (byte)0x01);
    }

    protected void envFireExtinguisherOff(int device_id)
    {
        _dev.setStatus(DEV_FIRE_EXTINGUISHER, device_id, (byte)0x02);
    }

    protected byte envReadLiftCaller(int device_id)
    {
        return (byte)(_dev.getStatus(DEV_LIFT_CONTROLLER, device_id) & 0x1F);
    }

    protected boolean envHasBadVolume(int device_id)
    {
        return 0 != (_dev.getStatus(DEV_VOLUME_SENSOR, device_id) & 0x01);
    }
}
