package BSS.Simulator;

import BSS.*;
import BSS.Controllers.*;
import BSS.Devices.*;

import java.util.*;
import java.io.IOException;

import org.w3c.dom.*;
import org.xml.sax.SAXException;

import javax.xml.parsers.*;

public class Runner implements EnvironmentCodes
{
    MainController _MC;
    Integer _MC_ID;

    Map _rooms; // Integer -> Room
    final static Map _room_types = new HashMap();

    Map _doors; // Integer -> Entry
    Map _gateways; // Integer -> Entry
    Map _passages; // Integer -> Entry

    Map _cashmachines; // Integer -> Service
    Map _cashdesks; // Integer -> Service
    Map _parkings; // Integer -> Service

    Map _configuration; // Integer -> HashMap<String -> Object>
    Map _devices; // Integer -> BaseDevice
    Map _controllers; // Integer -> Controller
    MessageCenter _communicator;
    Logger _logger;

    LinkedList _scenario;
    final static Map _event_codes = new HashMap();
    final static Map _sms_codes = new HashMap();
    final static Map _status_device_codes = new HashMap();
    final static Map _data_device_codes = new HashMap();

    static
    {
        _room_types.put("public", new Integer(Room.PUBLIC_PLACE));
        _room_types.put("clients", new Integer(Room.CLIENTS_ROOM));
        _room_types.put("staff", new Integer(Room.STAFF_ROOM));
        _room_types.put("storage", new Integer(Room.STORAGE));
        _room_types.put("lift", new Integer(Room.LIFT));

        _event_codes.put("push", new Integer(EVENT_PUSH));
        _event_codes.put("motion", new Integer(EVENT_MOTION));
        _event_codes.put("break", new Integer(EVENT_GLASS_BREAK));
        _event_codes.put("call", new Integer(EVENT_LIFT_CALL));
        _event_codes.put("robbery", new Integer(EVENT_ROBBERY_BTN));
        _event_codes.put("on", new Integer(EVENT_ALARM_ON_BTN));
        _event_codes.put("off", new Integer(EVENT_ALARM_OFF_BTN));

        _sms_codes.put("on", new Integer(SMS_SEC_ON));
        _sms_codes.put("off", new Integer(SMS_SEC_OFF));
        _sms_codes.put("smoke", new Integer(SMS_SMOKE_IN_ROOM));
        _sms_codes.put("lock", new Integer(SMS_LOCK_ENTRY));

        _status_device_codes.put("contact", new Integer(DEV_CONTACT_SENSOR));
        _status_device_codes.put("smartcard", new Integer(DEV_SMARTCARD_DRIVE));
        _status_device_codes.put("photo", new Integer(DEV_PHOTO_SENSOR));
        _status_device_codes.put("metal", new Integer(DEV_METAL_SENSOR));
        _status_device_codes.put("fire", new Integer(DEV_FIRE_SENSOR));
        _status_device_codes.put("lift", new Integer(DEV_LIFT_CONTROLLER));
        _status_device_codes.put("volume", new Integer(DEV_VOLUME_SENSOR));

        _data_device_codes.put("smartcard", new Integer(DEV_SMARTCARD_DRIVE));
        _data_device_codes.put("keyboard", new Integer(DEV_KEYBOARD));
    }

    //--- Public section ---
    public Runner(Logger logger)
    {
        _logger = logger;
    }

    public void loadConfiguration(String filename) throws ParserConfigurationException, IOException, SAXException
    {
        log(1, "Started loading configuration");

        List objects = getChildElements(parse_XML_File(filename).getDocumentElement());

        // Build architecture and find cross-relations
        log(2, "Building architecture objects and determining cross-relations...");

        _rooms = new TreeMap();
        _doors = new TreeMap();
        _gateways = new TreeMap();
        _passages = new TreeMap();
        _cashmachines = new TreeMap();
        _cashdesks = new TreeMap();
        _parkings = new TreeMap();

        Map room_entry_ids = new TreeMap(); // Integer -> TreeSet<Integer>
        Map entry_room_ids = new TreeMap(); // Integer -> TreeSet<Integer>
        Map service_room_id = new TreeMap(); // Integer -> Integer

        for (Iterator it = objects.iterator(); it.hasNext();)
        {
            Element object = (Element)it.next();
            String o_name = object.getNodeName();

            if ("maincontroller".equals(o_name))
            {
                _MC_ID = parseAttrInt(object, "id");
            }
            else if ("room".equals(o_name))
            {
                Integer roomID = parseAttrInt(object, "id");
                int type = parseRoomType(roomID.intValue(), parseAttrStr(object, "type"));
                int floor = parseAttrInt(object, "floor").intValue();
                Room room = new Room(roomID.intValue(), type, floor);
                _rooms.put(roomID, room);
                log(3, room.toString() + " floor=" + floor);
            }
            else if ("door".equals(o_name))
            {
                Integer doorID = parseAttrInt(object, "id");
                parseEntryRooms(object, doorID, entry_room_ids, room_entry_ids);
                Entry door = new Entry(doorID.intValue(), Entry.DOOR);
                _doors.put(doorID, door);
                log(3, door.toString());
            }
            else if ("gateway".equals(o_name))
            {
                Integer gatewayID = parseAttrInt(object, "id");
                parseEntryRooms(object, gatewayID, entry_room_ids, room_entry_ids);
                Entry gateway = new Entry(gatewayID.intValue(), Entry.GATEWAY);
                _gateways.put(gatewayID, gateway);
                log(3, gateway.toString());
            }
            else if ("passage".equals(o_name))
            {
                Integer passageID = parseAttrInt(object, "id");
                parseEntryRooms(object, passageID, entry_room_ids, room_entry_ids);
                Entry passage = new Entry(passageID.intValue(), Entry.PASSAGE);
                _passages.put(passageID, passage);
                log(3, passage.toString());
            }
            else if ("cashmachine".equals(o_name))
            {
                Integer cmID = parseAttrInt(object, "id");
                Integer roomID = parseAttrInt(object, "room");
                service_room_id.put(cmID, roomID);
                Service cm = new Service(cmID.intValue(), Service.CASH_MACHINE);
                _cashmachines.put(cmID, cm);
                log(3, cm.toString());
            }
            else if ("cashdesk".equals(o_name))
            {
                Integer cdID = parseAttrInt(object, "id");
                Integer roomID = parseAttrInt(object, "room");
                service_room_id.put(cdID, roomID);
                Service cd = new Service(cdID.intValue(), Service.CASH_DESK);
                _cashdesks.put(cdID, cd);
                log(3, cd.toString());
            }
            else if ("parking".equals(o_name))
            {
                Integer parkingID = parseAttrInt(object, "id");
                Integer roomID = parseAttrInt(object, "room");
                service_room_id.put(parkingID, roomID);
                Service parking = new Service(parkingID.intValue(), Service.PARKING);
                _parkings.put(parkingID, parking);
                log(3, parking.toString());
            }
        }

        Map all_entries = new TreeMap(); // Integer -> Entry
        all_entries.putAll(_doors);
        all_entries.putAll(_gateways);
        all_entries.putAll(_passages);

        Map services = new TreeMap();
        services.putAll(_cashmachines);
        services.putAll(_cashdesks);
        services.putAll(_parkings);

        // Set relations
        log(2, "Setuping cross-relations...");

        for (Iterator it = all_entries.entrySet().iterator(); it.hasNext();)
        {
            Map.Entry pair = (Map.Entry) it.next();
            Integer entryID = (Integer)pair.getKey();
            Entry entry = (Entry)pair.getValue();
            Set rooms = getMapValuesByKeySet((Set)entry_room_ids.get(entryID), _rooms);
            entry.setRooms(rooms);
            log(3, entry.toString() + " rooms=" + rooms);
        }

        for (Iterator it = _rooms.entrySet().iterator(); it.hasNext();)
        {
            Map.Entry pair = (Map.Entry) it.next();
            Integer roomID = (Integer)pair.getKey();
            Room room = (Room)pair.getValue();
            Set entries = getMapValuesByKeySet((Set)room_entry_ids.get(roomID), all_entries);
            room.setEntries(entries);
            log(3, room.toString() + " entries=" + entries);
        }

        for (Iterator it = services.entrySet().iterator(); it.hasNext();)
        {
            Map.Entry pair = (Map.Entry) it.next();
            Integer serviceID = (Integer)pair.getKey();
            Service service = (Service)pair.getValue();
            Room room = (Room)_rooms.get(service_room_id.get(serviceID));
            service.setRoom(room);
            log(3, service.toString() + " room=" + room);
        }

        // Create configuration, communicator, devices and controllers
        log(2, "Creating configuration, communicator, devices and controllers...");

        _configuration = new TreeMap();
        _devices = new TreeMap();
        _controllers = new TreeMap();
        _communicator = new MessageCenter(_logger, _MC_ID.intValue());

        for (Iterator it = objects.iterator(); it.hasNext();)
        {
            Element object = (Element)it.next();
            String o_name = object.getNodeName();

            if ("maincontroller".equals(o_name))
            {
                Map conf = new HashMap();
                conf.put("Rooms", _rooms);
                conf.put("Doors", _doors);
                conf.put("Gateways", _gateways);
                conf.put("Passages", _passages);
                conf.put("CashMachines", _cashmachines);
                conf.put("CashDesks", _cashdesks);
                conf.put("Parkings", _parkings);
                _configuration.put(_MC_ID, conf);

                BaseDevice device = new MainControllerDevice(_MC_ID.intValue(), _logger, conf, _communicator);
                _devices.put(_MC_ID, device);
                _MC = new MainController(_MC_ID.intValue(), _logger, device);
                _controllers.put(_MC_ID, _MC);
                device.setController(_MC);

                log(3, _MC.toString() + " configuration=" + conf);
            }
            if ("room".equals(o_name))
            {
                Map conf = new HashMap();
                Integer roomID = parseAttrInt(object, "id");
                conf.put("MC", _MC_ID);
                conf.put("ActiveFS", new Boolean("active".equals(parseAttrStr(object, "fs"))));
                conf.put("Doors", selectEntriesByType(Entry.DOOR, (Set)room_entry_ids.get(roomID), all_entries));
                conf.put("Gateways", selectEntriesByType(Entry.GATEWAY, (Set)room_entry_ids.get(roomID), all_entries));
                conf.put("Passages", selectEntriesByType(Entry.PASSAGE, (Set)room_entry_ids.get(roomID), all_entries));
                Room room = (Room)_rooms.get(roomID);
                conf.put("Type", new Integer(room.getType()));
                if (Room.LIFT == room.getType())
                {
                    conf.put("LowerFloor", new Integer(room.getFloor()));
                }
                _configuration.put(roomID, conf);

                BaseDevice device = new RoomDevice(roomID.intValue(), _logger, conf, _communicator);
                _devices.put(roomID, device);
                RoomController roomc = new RoomController(roomID.intValue(), _logger, device);
                _controllers.put(roomID, roomc);
                device.setController(roomc);

                log(3, roomc.toString() + " configuration=" + conf);
            }
            else if ("door".equals(o_name))
            {
                Map conf = new HashMap();
                Integer doorID = parseAttrInt(object, "id");
                conf.put("MC", _MC_ID);
                conf.put("PIN", parse_PIN(object));
                _configuration.put(doorID, conf);

                BaseDevice device = new DoorDevice(doorID.intValue(), _logger, conf, _communicator);
                _devices.put(doorID, device);
                DoorController door = new DoorController(doorID.intValue(), _logger, device);
                _controllers.put(doorID, door);
                device.setController(door);

                log(3, door.toString() + " configuration=" + conf);
            }
            else if ("gateway".equals(o_name))
            {
                Map conf = new HashMap();
                Integer gatewayID = parseAttrInt(object, "id");
                conf.put("MC", _MC_ID);
                boolean has_photocell = "yes".equals(parseAttrStr(object, "photo"));
                conf.put("PhotoCell", new Boolean(has_photocell));
                _configuration.put(gatewayID, conf);

                BaseDevice device = new GatewayDevice(gatewayID.intValue(), _logger, conf, _communicator, has_photocell);
                _devices.put(gatewayID, device);
                GatewayController gateway = new GatewayController(gatewayID.intValue(), _logger, device);
                _controllers.put(gatewayID, gateway);
                device.setController(gateway);

                log(3, gateway.toString() + " configuration=" + conf);
            }
            else if ("cashmachine".equals(o_name))
            {
                Map conf = new HashMap();
                Integer cmID = parseAttrInt(object, "id");
                conf.put("MC", _MC_ID);
                _configuration.put(cmID, conf);

                BaseDevice device = new CashMachineDevice(cmID.intValue(), _logger, conf, _communicator);
                _devices.put(cmID, device);
                CashMachineController cashmachine = new CashMachineController(cmID.intValue(), _logger, device);
                _controllers.put(cmID, cashmachine);
                device.setController(cashmachine);

                log(3, cashmachine.toString() + " configuration=" + conf);
            }
            else if ("cashdesk".equals(o_name))
            {
                Map conf = new HashMap();
                Integer cdID = parseAttrInt(object, "id");
                conf.put("MC", _MC_ID);
                _configuration.put(cdID, conf);

                BaseDevice device = new CashDeskDevice(cdID.intValue(), _logger, conf, _communicator);
                _devices.put(cdID, device);
                CashDeskController cashdesk = new CashDeskController(cdID.intValue(), _logger, device);
                _controllers.put(cdID, cashdesk);
                device.setController(cashdesk);

                log(3, cashdesk.toString() + " configuration=" + conf);
            }
            else if ("parking".equals(o_name))
            {
                Map conf = new HashMap();
                Integer parkingID = parseAttrInt(object, "id");
                conf.put("MC", _MC_ID);
                _configuration.put(parkingID, conf);

                BaseDevice device = new ParkingDevice(parkingID.intValue(), _logger, conf, _communicator);
                _devices.put(parkingID, device);
                ParkingController parking = new ParkingController(parkingID.intValue(), _logger, device);
                _controllers.put(parkingID, parking);
                device.setController(parking);

                log(3, parking.toString() + " configuration=" + conf);
            }
        }
        _communicator.setHosts(_devices);

        log(1, "Finished loading configuration");
    }

    public void loadScenario(String filename) throws ParserConfigurationException, IOException, SAXException
    {
        log(1, "Started loading scenario");

        _scenario = new LinkedList();
        List instructions = getChildElements(parse_XML_File(filename).getDocumentElement());
        for (Iterator it = instructions.iterator(); it.hasNext(); )
        {
            Element instruction = (Element)it.next();;
            String i_name = instruction.getNodeName();

            if ("delay".equals(i_name))
            {
                int milliseconds = parseAttrInt(instruction, "milliseconds").intValue();
                _scenario.addLast(new DoDelay(milliseconds));
                log(3, ":DELAY " + milliseconds + "ms");
            }
            else if ("event".equals(i_name))
            {
                int device = parseAttrInt(instruction, "device").intValue();
                String code    = parseAttrStr(instruction, "code");
                Integer event_code = (Integer)_event_codes.get(code);
                if (null != event_code)
                {
                    _scenario.addLast(new RaiseEvent(device, event_code.intValue()));
                    log(3, ":EVENT `" + code + "` for controller#" + device);
                }
                else
                {
                    log(1, "Missed EVENT instruction with unknown event code `" + code + "`");
                }
            }
            else if ("sms".equals(i_name))
            {
                int object = parseAttrInt(instruction, "object").intValue();
                String action = parseAttrStr(instruction, "action");
                Integer sms_code = (Integer)_sms_codes.get(action);
                if (null != sms_code)
                {
                    _scenario.addLast(new SMS("BSS", "Chief", object, sms_code.intValue()));
                    log(3, ":SMS action `" + action + "` for controller#" + object);
                }
                else
                {
                    log(1, "Missed SMS instruction with unknown action code `" + action + "`");
                }
            }
            else if ("status".equals(i_name))
            {
                int device = parseAttrInt(instruction, "device").intValue();
                int number = parseAttrInt(instruction, "number").intValue();
                byte value = (byte)parseAttrInt(instruction, "value").intValue();
                String type = parseAttrStr(instruction, "type");
                Integer device_type = (Integer)_status_device_codes.get(type);
                if (null != device_type)
                {
                    _scenario.addLast(new SetStatus(device, device_type.intValue(), number, value));
                    log(3, ":STATUS of device type `" + type + "` for controller #" + device + "[" + number + "]=" + value);
                }
                else
                {
                    log(1, "Missed STATUS instruction with unknown device type `" + type + "`");
                }
            }
            else if ("data".equals(i_name))
            {
                int device = parseAttrInt(instruction, "device").intValue();
                int number = parseAttrInt(instruction, "number").intValue();
                String value = instruction.getNodeValue();
                String type = parseAttrStr(instruction, "type");
                Integer device_type = (Integer)_data_device_codes.get(type);
                if (null != device_type)
                {
                    _scenario.addLast(new SetDataBuffer(device, device_type.intValue(), number, value.getBytes()));
                    log(3, ":DATA of device type `" + type + "` for controller #" + device + "[" + number + "]=" + value);
                }
                else
                {
                    log(1, "Missed DATA instruction with unknown device type `" + type + "`");
                }
            }
        }

        log(1, "Finished loading scenario");
    }

    public void init()
    {
        log(1, "Started initializing controllers");

        for (Iterator iterator = _controllers.values().iterator(); iterator.hasNext();)
        {
            ((Controller) iterator.next()).init();
        }

        log(1, "Finished initializing controllers");
    }

    public void runScenario()
    {
        log(1, "Started running scenario");

        for (Iterator iterator = _scenario.iterator(); iterator.hasNext();)
        {
            Object step = iterator.next();
            if (step instanceof DoDelay)
            {
                log(2, "!Processing DELAY");
                DoDelay do_delay = (DoDelay)step;
                long end_time = System.currentTimeMillis() + do_delay.milliseconds;
                do
                {
                    for (Iterator it = _controllers.values().iterator(); it.hasNext();)
                    {
                        Controller controller = (Controller) it.next();
                        controller.run();
                    }
                }
                while (System.currentTimeMillis() <= end_time);
            }
            else if (step instanceof RaiseEvent)
            {
                log(2, "!Processing EVENT");
                RaiseEvent raise_event = (RaiseEvent)step;
                Host host = (Host)_devices.get(new Integer(raise_event.device));
                if (null != host)
                {
                    host.interrupt(raise_event.code);
                }
                else
                {
                    log(1, "Unknown device #" + raise_event.device);
                }
            }
            else if (step instanceof SMS)
            {
                log(2, "!Processing SMS");
                SMS sms = (SMS)step;
                _communicator.sendSMS(sms);
            }
            else if (step instanceof SetStatus)
            {
                log(2, "!Processing STATUS");
                SetStatus set_status = (SetStatus)step;
                BaseDevice device = (BaseDevice)_devices.get(new Integer(set_status.device));
                if (null != device)
                {
                    device.updateStatus(set_status.type, set_status.number, set_status.value);
                }
                else
                {
                    log(1, "Unknown device #" + set_status.device);
                }
            }
            else if (step instanceof SetDataBuffer)
            {
                log(2, "Processing DATA...");
                SetDataBuffer set_data = (SetDataBuffer)step;
                BaseDevice device = (BaseDevice)_devices.get(new Integer(set_data.device));
                if (null != device)
                {
                    device.updateData(set_data.type, set_data.number, set_data.value);
                }
                else
                {
                    log(1, "Unknown device #" + set_data.device);
                }
            }
        }

        log(1, "Finished running scenario");
    }

    //--- Methods for private usage ---
    private void log(int level, String message)
    {
        _logger.log(level, message);
    }

    private Set getMapValuesByKeySet(Set keys, Map map)
    {
        Set values = new HashSet();
        for (Iterator it = keys.iterator(); it.hasNext();)
        {
            values.add(map.get(it.next()));
        }
        return values;
    }

    private Set selectEntriesByType(int type, Set entry_ids, Map entries)
    {
        Set selection = new HashSet();
        for (Iterator it = entry_ids.iterator(); it.hasNext();)
        {
            Integer entryID = (Integer) it.next();
            Entry entry = (Entry)entries.get(entryID);
            if (type == entry.getType())
            {
                selection.add(entryID);
            }
        }
        return selection;
    }

    private int parseRoomType(int room_id, String type)
    {
        Integer room_type = (Integer)_room_types.get(type);
        if (null == room_type)
        {
            room_type = new Integer(Room.PUBLIC_PLACE);
            log(1, "Unknown type `" + type + "` of room#" + room_id + " set by default to public.");
        }
        return room_type.intValue();
    }

    //--- DOM parsing methods ---
    private Document parse_XML_File(String filename) throws ParserConfigurationException, IOException, SAXException
    {
        log(1, "Parsing XML file `" + filename + "` ...");
        Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(filename);
        log(1, "XML parsed.");
        return doc;
    }

    private List getChildElements(Element element)
    {
        List elements = new ArrayList();

        NodeList nodes = element.getChildNodes();
        int len = nodes.getLength();
        for (int i = 0; i < len; i++)
        {
            Node node = nodes.item(i);
            if (node.getNodeType() == Node.ELEMENT_NODE)
            {
                elements.add(node);
            }
        }

        return elements;
    }

    private String parseAttrStr(Element element, String attr_name)
    {
        return element.getAttribute(attr_name);
    }

    private Integer parseAttrInt(Element element, String attr_name)
    {
        return new Integer(element.getAttribute(attr_name));
    }

    private void parseEntryRooms(Element object, Integer entryID, Map entry_rooms, Map room_entries)
    {
        Set rooms = new TreeSet();
        for (Iterator it = getChildElements(object).iterator(); it.hasNext();)
        {
            Element prop = (Element) it.next();
            if ("adjacent".equals(prop.getNodeName()))
            {
                Integer roomID = parseAttrInt(prop, "room");
                rooms.add(roomID);
                Set entries = (Set)room_entries.get(roomID);
                if (null == entries)
                {
                    entries = new TreeSet();
                    room_entries.put(roomID, entries);
                }
                entries.add(entryID);
            }
        }
        entry_rooms.put(entryID, rooms);
    }

    private String parse_PIN(Element object)
    {
        String pin = "";
        for (Iterator it = getChildElements(object).iterator(); it.hasNext();)
        {
            Element prop = (Element)it.next();
            if ("pin".equals(prop.getNodeName()))
            {
                pin = prop.getFirstChild().getNodeValue();
                break;
            }
        }
        return pin;
    }

    //--- Runner ---
    public static void main(String[] args) throws Exception
    {
        int log_level = Integer.parseInt(System.getProperty("LOG_LEVEL", "2"));
        String timestamp_template = System.getProperty("TIMESTAMP", "hh:mm:ss.SSS");
        boolean use_xml_logger = "xml".equals(System.getProperty("LOGGER", "simple"));

        String config_file = System.getProperty("CONFIG");
        if (null == config_file)
        {
            System.err.println("Configuration XML-file not set.");
            System.err.println("Usage: java -DCONFIG=<configuration file> -DSCENARIO=<scenario file> [-DLOG_LEVEL=<non-negative number>] [-DTIMESTAMP=<timestamp template>] [-DLOGGER=simple|xml] BSS.Simulator.TestRunner");
            return;
        }
        String scenario_file = System.getProperty("SCENARIO");
        if (null == scenario_file)
        {
            System.err.println("Scenario XML-file not set.");
            System.err.println("Usage: java -DCONFIG=<configuration file> -DSCENARIO=<scenario file> [-DLOG_LEVEL=<non-negative number>] [-DTIMESTAMP=<timestamp template>] [-DLOGGER=simple|xml] BSS.Simulator.TestRunner");
            return;
        }

        Logger logger;
        if (use_xml_logger)
        {
            logger = new XML_Logger(log_level, timestamp_template, System.out);
        }
        else
        {
            logger = new SimpleLogger(log_level, timestamp_template, System.out);
        }
        logger.start_log();

        Runner env = new Runner(logger);
        env.loadConfiguration(config_file);
        env.loadScenario(scenario_file);
        env.init();
        env.runScenario();

        logger.end_log();
    }

    //--- Instruction classes ---
    class DoDelay
    {
        public int milliseconds;

        public DoDelay(int milliseconds)
        {
            this.milliseconds = milliseconds;
        }
    }

    class RaiseEvent
    {
        public int device;
        public int code;

        public RaiseEvent(int device, int code)
        {
            this.device = device;
            this.code = code;
        }
    }

    class SetStatus
    {
        public int device;
        public int type;
        public int number;
        public byte value;

        public SetStatus(int device, int type, int number, byte value)
        {
            this.device = device;
            this.type = type;
            this.number = number;
            this.value = value;
        }
    }

    class SetDataBuffer
    {
        public int device;
        public int type;
        public int number;
        public byte[] value;

        public SetDataBuffer(int device, int type, int number, byte[] value)
        {
            this.device = device;
            this.type = type;
            this.number = number;
            this.value = value;
        }
    }
}
